@@ -1,12 +1,13 @@
 #  GTK+ based frontend to software-properties
 #
-#  Copyright (c) 2004-2018 Canonical Ltd.
+#  Copyright (c) 2004-2020 Canonical Ltd.
 #                2004-2005 Michiel Sikkes
 #
 #  Author: Michiel Sikkes <michiel@eyesopened.nl>
 #          Michael Vogt <mvo@debian.org>
 #          Sebastian Heinlein <glatzor@ubuntu.com>
 #          Andrea Azzarone <andrea.azzarone@canonical.com>
+#          Olivier Tilloy <olivier.tilloy@canonical.com>
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU General Public License as
@@ -27,16 +28,13 @@ from __future__ import absolute_import, print_function
 
 import apt
 import apt_pkg
-import aptsources.distro
-from datetime import datetime
-import distro_info
 import dbus
+import dbus.glib
+import dbus.mainloop.glib
 from gettext import gettext as _
 import gettext
 import os
 import subprocess
-from aptdaemon import client
-from aptdaemon.errors import NotAuthorizedError, TransactionFailed
 import logging
 import threading
 import sys
@@ -44,6 +42,8 @@ import sys
 import gi
 gi.require_version("Gdk", "3.0")
 gi.require_version("Gtk", "3.0")
+gi.require_version("PackageKitGlib", "1.0")
+from gi.repository import PackageKitGlib as packagekit
 from gi.repository import GObject, Gdk, Gtk, Gio, GLib
 
 from .SimpleGtkbuilderApp import SimpleGtkbuilderApp
@@ -52,16 +52,15 @@ from .DialogMirror import DialogMirror
 from .DialogEdit import DialogEdit
 from .DialogCacheOutdated import DialogCacheOutdated
 from .DialogAddSourcesList import DialogAddSourcesList
-from .DialogLivepatchError import DialogLivepatchError
-from .DialogAuth import DialogAuth
+from .LivepatchPage import LivepatchPage
 
 import softwareproperties
-from softwareproperties.GoaAuth import GoaAuth
 import softwareproperties.distro
 from softwareproperties.SoftwareProperties import SoftwareProperties
 import softwareproperties.SoftwareProperties
 
 from UbuntuDrivers import detect
+dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
 
 if GLib.pyglib_version < (3, 9, 1):
     GLib.threads_init()
@@ -85,7 +84,51 @@ RESPONSE_ADD = 2
     STORE_VISIBLE
 ) = list(range(5))
 
-LIVEPATCH_TIMEOUT = 1200
+# rows for the updates subscription combobox
+(
+    UPDATES_ALL,
+    UPDATES_RECOMMENDED,
+    UPDATES_SECURITY,
+    UPDATES_CUSTOM,
+) = range(4)
+
+# names of the pockets
+POCKET_SECURITY = "security"
+POCKET_UPDATES = "updates"
+POCKET_BACKPORTS = "backports"
+
+
+class UpdateSubscriptions(object):
+    security = False
+    updates = False
+    backports = False
+    other = False
+    inconsistent = False
+
+    def get_status(self):
+        if self.other:
+            return UPDATES_CUSTOM
+        if self.security and self.updates and self.backports:
+            return UPDATES_ALL
+        elif self.security and self.updates and not self.backports:
+            return UPDATES_RECOMMENDED
+        elif self.security and not self.updates and not self.backports:
+            return UPDATES_SECURITY
+        return UPDATES_CUSTOM
+
+
+def maybe_log_authentication_canceled_error(dbus_exception):
+    """Log an error if the DBus exception was raised because the user canceled
+    the authentication prompt shown when changing apt sources/configuration.
+
+    Return True if it is, so that calling code can do further processing,
+    False otherwise.
+    """
+    if dbus_exception._dbus_error_name == \
+      'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
+        logging.error("Authentication canceled, changes have not been saved")
+        return True
+    return False
 
 
 def error(parent_window, summary, msg):
@@ -185,7 +228,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         self.handlers = {}
 
         self.apt_cache = {}
-        self.apt_client = client.AptClient()
+        self.pk_task = None
 
         # Put some life into the user interface:
         self.init_auto_update()
@@ -351,29 +394,28 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             self.vbox_dist_comps.add(checkbox)
             checkbox.show()
 
-        # Setup the checkbuttons for the child repos / updates
-        for checkbutton in self.vbox_updates.get_children():
-            self.vbox_updates.remove(checkbutton)
+        # Setup the checkbuttons for the child repos
         for checkbutton in self.dev_box.get_children():
             self.dev_box.remove(checkbutton)
-        if len(self.distro.source_template.children) < 1:
-            self.frame_children.hide()
         for template in self.distro.source_template.children:
             # Do not show source entries in there
             if template.type == "deb-src":
                 continue
-
-            checkbox = Gtk.CheckButton(label="%s (%s)" % (template.description,
-                                                          template.name))
-            checkbox.template = template
-            self.handlers[checkbox] = checkbox.connect("toggled",
-                                                   self.on_checkbutton_child_toggled,
-                                                   template)
             if "proposed" in template.name:
+                checkbox = Gtk.CheckButton(label="%s (%s)" % (template.description,
+                                                              template.name))
+                checkbox.template = template
+                self.handlers[checkbox] = \
+                    checkbox.connect("toggled",
+                                     self.on_checkbutton_child_toggled,
+                                     template)
                 self.dev_box.add(checkbox)
-            else:
-                self.vbox_updates.add(checkbox)
-            checkbox.show()
+                checkbox.show()
+
+        # Setup the combo box for updates subscriptions
+        combobox = self.combobox_updates_subscription
+        self.handlers[combobox] = combobox.connect(
+            "changed", self.on_combobox_updates_subscription_changed)
 
         # setup the server chooser
         cell = Gtk.CellRendererText()
@@ -401,12 +443,6 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         """Fill the distro user interface with life"""
         self.block_handlers()
 
-        # Enable or disable the child source checkbuttons
-        for checkbox in self.vbox_updates.get_children():
-            (active, inconsistent) = self.get_comp_child_state(checkbox.template)
-            checkbox.set_active(active)
-            checkbox.set_inconsistent(inconsistent)
-
         # Enable or disable the child source checkbuttons
         for checkbox in self.dev_box.get_children():
             (active, inconsistent) = self.get_comp_child_state(checkbox.template)
@@ -420,14 +456,48 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             checkbox.set_inconsistent(inconsistent)
             checkbox.set_active(active)
 
+        # Refresh the combo box for updates subscriptions
+        subscriptions = UpdateSubscriptions()
+        for template in self.distro.source_template.children:
+            # Do not show source entries in there
+            if template.type == "deb-src":
+                continue
+            pocket = template.name.split('-')[-1]
+            (active, inconsistent) = self.get_comp_child_state(template)
+            subscriptions.inconsistent |= inconsistent
+            if active:
+                if pocket == POCKET_SECURITY:
+                    subscriptions.security = True
+                elif pocket == POCKET_UPDATES:
+                    subscriptions.updates = True
+                elif pocket == POCKET_BACKPORTS:
+                    subscriptions.backports = True
+                else:
+                    subscriptions.other = True
+
+        status = subscriptions.get_status()
+        combobox = self.combobox_updates_subscription
+        model = combobox.get_model()
+        if status == UPDATES_CUSTOM:
+            try:
+                model.get_iter_from_string(str(UPDATES_CUSTOM))
+            except ValueError:
+                model.append([_("Custom")])
+        else:
+            try:
+                model.remove(model.get_iter_from_string(str(UPDATES_CUSTOM)))
+            except ValueError:
+                pass
+        combobox.set_active(status)
+        combobox.set_sensitive(not subscriptions.inconsistent)
+
         # If no components are enabled there will be no need for updates
         # and source code
         if len(self.distro.enabled_comps) < 1:
-            self.vbox_updates.set_sensitive(False)
+            self.combobox_updates_subscription.set_sensitive(False)
             self.dev_box.set_sensitive(False)
             self.checkbutton_source_code.set_sensitive(False)
         else:
-            self.vbox_updates.set_sensitive(True)
             self.dev_box.set_sensitive(True)
             self.checkbutton_source_code.set_sensitive(True)
 
@@ -514,9 +584,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             if state != self.get_update_automation_level():
                 self.backend.SetUpdateAutomationLevel(state)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
-
+            if maybe_log_authentication_canceled_error(e):
                 combo_handler = self.handlers[self.combobox_security_updates]
                 self.combobox_security_updates.handler_block(combo_handler)
                 self.set_security_update_level()
@@ -541,15 +609,52 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.SetReleaseUpgradesPolicy(i)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
-
+            if maybe_log_authentication_canceled_error(e):
                 combo_handler = self.handlers[self.combobox_release_upgrades]
                 self.combobox_release_upgrades.handler_block(combo_handler)
                 i = self.get_release_upgrades_policy()
                 self.combobox_release_upgrades.set_active(i)
                 self.combobox_release_upgrades.handler_unblock(combo_handler)
 
+    def on_combobox_updates_subscription_changed(self, combobox):
+        index = combobox.get_active()
+        if index != UPDATES_CUSTOM:
+            model = combobox.get_model()
+            try:
+                model.remove(model.get_iter_from_string(str(UPDATES_CUSTOM)))
+            except ValueError:
+                pass
+
+        def _toggle_child_source(pocket, enabled):
+            source = "{}-{}".format(self.distro.codename, pocket)
+            if enabled:
+                logging.info("enabling {}".format(source))
+                self.backend.EnableChildSource(source)
+            else:
+                logging.info("disabling {}".format(source))
+                self.backend.DisableChildSource(source)
+
+        try:
+            if index == UPDATES_ALL:
+                _toggle_child_source(POCKET_SECURITY, True)
+                _toggle_child_source(POCKET_UPDATES, True)
+                _toggle_child_source(POCKET_BACKPORTS, True)
+            elif index == UPDATES_RECOMMENDED:
+                _toggle_child_source(POCKET_SECURITY, True)
+                _toggle_child_source(POCKET_UPDATES, True)
+                _toggle_child_source(POCKET_BACKPORTS, False)
+            elif index == UPDATES_SECURITY:
+                _toggle_child_source(POCKET_SECURITY, True)
+                _toggle_child_source(POCKET_UPDATES, False)
+                _toggle_child_source(POCKET_BACKPORTS, False)
+            for template in self.distro.source_template.children:
+                if template.type == "deb-src":
+                    continue
+                pocket = template.name.split('-')[-1]
+                if pocket not in (POCKET_SECURITY, POCKET_UPDATES, POCKET_BACKPORTS):
+                    _toggle_child_source(pocket, False)
+        except dbus.DBusException as e:
+            maybe_log_authentication_canceled_error(e)
 
     def on_combobox_server_changed(self, combobox):
         """
@@ -577,8 +682,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                 self.active_server = combobox.get_active()
                 self.backend.ChangeMainDownloadServer(uri)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
         # mvo: is this still needed?
         #else:
@@ -595,8 +699,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.DisableComponent(comp)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_checkbutton_child_toggled(self, checkbutton, template):
         """
@@ -608,8 +711,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.EnableChildSource(template.name)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_checkbutton_source_code_toggled(self, checkbutton):
         """ Disable or enable the source code for all sources """
@@ -619,8 +721,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.DisableSourceCodeSources()
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_checkbutton_popcon_toggled(self, widget):
         """ The user clicked on the popcon paritipcation button """
@@ -630,8 +731,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             try:
                 self.backend.SetPopconPariticipation(widget.get_active())
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     def open_file(self, file):
         """Show a confirmation for adding the channels of the specified file"""
@@ -649,8 +749,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                     for source in new_sources:
                         self.backend.AddSourceFromLine(str(source))
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_sources_drag_data_received(self, widget, context, x, y,
                                       selection, target_type, timestamp):
@@ -754,8 +853,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.ToggleSourceUse(str(source_entry))
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def init_keys(self):
         """Setup the user interface parts needed for the key handling"""
@@ -816,8 +914,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.Revert()
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
         if self.settings:
             self.settings.set_int("regular-auto-launch-interval", self.initial_auto_launch)
         self.show_auto_update_level()
@@ -901,9 +998,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             try:
                 self.backend.SetUpdateInterval(value)
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
-
+                if maybe_log_authentication_canceled_error(e):
                     update_days = self.get_update_interval()
                     combo_handler = self.handlers[self.combobox_update_interval]
                     for key in self.combobox_interval_mapping:
@@ -922,8 +1017,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             try:
                 self.backend.AddSourceFromLine(line)
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     def on_edit_clicked(self, widget):
         """Show a dialog to edit an ISV source"""
@@ -939,8 +1033,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                 self.backend.ReplaceSourceEntry(str(old_source_entry),
                                                 str(dialog.new_source_entry))
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     # FIXME:outstanding from merge
     def on_isv_source_activated(self, treeview, path, column):
@@ -980,8 +1073,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             try:
                 self.backend.RemoveSource(str(source_entry))
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     def add_key_clicked(self, widget):
         """Provide a file chooser that allows to add the gnupg of a trusted
@@ -1004,8 +1096,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                           _("The selected file may not be a GPG key file "
                             "or it might be corrupt."))
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     def remove_key_clicked(self, widget):
         """Remove a trusted software vendor key"""
@@ -1015,22 +1106,20 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             return
         key = model.get_value(a_iter, 0)
         try:
-            if not self.backend.RemoveKey(key[:8]):
+            if not self.backend.RemoveKey(key[:16]):
                 error(self.main,
                       _("Error removing the key"),
                       _("The key you selected could not be removed. "
                         "Please report this as a bug."))
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_restore_clicked(self, widget):
         """Restore the original keys"""
         try:
             self.backend.UpdateKeys()
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_delete_event(self, widget, args):
         """Close the window if requested"""
@@ -1045,7 +1134,9 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             d = DialogCacheOutdated(self.window_main,
                                     self.datadir)
             d.run()
-        if self.waiting_livepatch_response:
+
+        self.quit_when_livepatch_responds = False
+        if self.livepatch_page.waiting_livepatch_response:
             self.quit_when_livepatch_responds = True
             self.hide()
         else:
@@ -1056,10 +1147,9 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.AddCdromSource()
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
-    def on_driver_changes_progress(self, transaction, progress):
+    def on_driver_changes_progress(self, progress, ptype, data=None):
         #print(progress)
         self.button_driver_revert.set_visible(False)
         self.button_driver_apply.set_visible(False)
@@ -1069,39 +1159,36 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         self.progress_bar.set_visible(True)
 
         self.label_driver_action.set_label(_("Applying changes..."))
-        self.progress_bar.set_fraction(progress / 100.0)
+        if ptype == packagekit.ProgressType.PERCENTAGE:
+            prog_value = progress.get_property('percentage')
+            self.progress_bar.set_fraction(prog_value / 100.0)
 
-    def on_driver_changes_finish(self, transaction, exit_state):
-        self.progress_bar.set_visible(False)
-        self.clear_changes()
-        self.apt_cache = apt.Cache()
-        self.set_driver_action_status()
-        self.update_label_and_icons_from_status()
-        self.button_driver_revert.set_visible(True)
-        self.button_driver_apply.set_visible(True)
-        self.button_driver_cancel.set_visible(False)
-        self.scrolled_window_drivers.set_sensitive(True)
-
-    def on_driver_changes_error(self, transaction, error_code, error_details):
-        self.on_driver_changes_revert()
-        self.set_driver_action_status()
-        self.update_label_and_icons_from_status()
-        self.button_driver_revert.set_visible(True)
-        self.button_driver_apply.set_visible(True)
-        self.button_driver_cancel.set_visible(False)
-        self.scrolled_window_drivers.set_sensitive(True)
-
-    def on_driver_changes_cancellable_changed(self, transaction, cancellable):
-        self.button_driver_cancel.set_sensitive(cancellable)
+    def on_driver_changes_finish(self, source, result, installs_pending):
+        results = None
+        try:
+            results = self.pk_task.generic_finish(result)
+        except Exception as e:
+            self.on_driver_changes_revert()
+            error(self.window_main, _("Error while applying changes"), str(e))
+        if not installs_pending:
+            self.progress_bar.set_visible(False)
+            self.clear_changes()
+            self.apt_cache = apt.Cache()
+            self.set_driver_action_status()
+            self.update_label_and_icons_from_status()
+            self.button_driver_revert.set_visible(True)
+            self.button_driver_apply.set_visible(True)
+            self.button_driver_cancel.set_visible(False)
+            self.scrolled_window_drivers.set_sensitive(True)
 
     def on_driver_changes_apply(self, button):
-
+        self.pk_task = packagekit.Task()
         installs = []
         removals = []
 
         for pkg in self.driver_changes:
             if pkg.is_installed:
-                removals.append(pkg.shortname)
+                removals.append(self.get_package_id(pkg.installed))
                 # The main NVIDIA package is only a metapackage.
                 # We need to collect its dependencies, so that
                 # we can uninstall the driver properly.
@@ -1109,23 +1196,38 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                     for dep in get_dependencies(self.apt_cache, pkg.shortname, 'nvidia'):
                         dep_pkg = self.apt_cache[dep]
                         if dep_pkg.is_installed:
-                            removals.append(dep_pkg.shortname)
+                            removals.append(self.get_package_id(dep_pkg.installed))
             else:
-                installs.append(pkg.shortname)
+                installs.append(self.get_package_id(pkg.candidate))
 
+        self.cancellable = Gio.Cancellable()
         try:
-            self.transaction = self.apt_client.commit_packages(install=installs, remove=removals,
-                                                               reinstall=[], purge=[], upgrade=[], downgrade=[])
-            self.transaction.connect("progress-changed", self.on_driver_changes_progress)
-            self.transaction.connect("cancellable-changed", self.on_driver_changes_cancellable_changed)
-            self.transaction.connect("finished", self.on_driver_changes_finish)
-            self.transaction.connect("error", self.on_driver_changes_error)
-            self.transaction.set_debconf_frontend("gnome")
-            self.transaction.run()
+            if removals:
+                installs_pending = False
+                if installs:
+                    installs_pending = True
+                self.pk_task.remove_packages_async(removals,
+                            False,  # allow deps
+                            True,  # autoremove
+                            self.cancellable,  # cancellable
+                            self.on_driver_changes_progress,
+                            (None, ),  # progress data
+                            self.on_driver_changes_finish,  # callback ready
+                            installs_pending  # callback data
+                 )
+            if installs:
+                self.pk_task.install_packages_async(installs,
+                        self.cancellable,  # cancellable
+                        self.on_driver_changes_progress,
+                        (None, ),  # progress data
+                        self.on_driver_changes_finish,  # GAsyncReadyCallback
+                        False  # ready data
+                 )
+
             self.button_driver_revert.set_sensitive(False)
             self.button_driver_apply.set_sensitive(False)
             self.scrolled_window_drivers.set_sensitive(False)
-        except (NotAuthorizedError, TransactionFailed) as e:
+        except Exception as e:
             print("Warning: install transaction not completed successfully: {}".format(e))
 
     def on_driver_changes_revert(self, button_revert=None):
@@ -1145,7 +1247,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         self.button_driver_apply.set_sensitive(False)
 
     def on_driver_changes_cancel(self, button_cancel):
-        self.transaction.cancel()
+        self.cancellable.cancel()
         self.clear_changes()
 
     def on_driver_restart_clicked(self, button_restart):
@@ -1483,164 +1585,6 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         else:
             self.label_driver_action.set_label(_("No proprietary drivers are in use."))
 
-    #
-    # Livepatch
-    #
     def init_livepatch(self):
-        self.goa_auth = GoaAuth()
-        self.waiting_livepatch_response = False
-        self.quit_when_livepatch_responds = False
+        self.livepatch_page = LivepatchPage(self)
 
-        if not self.is_livepatch_supported():
-            self.grid_livepatch.set_visible(False)
-            return
-
-        self.checkbutton_livepatch.set_active(self.is_livepatch_enabled())
-        self.on_goa_auth_changed()
-
-        # hacky way to monitor if livepatch is enabled or not
-        file = Gio.File.new_for_path(path=self.LIVEPATCH_RUNNING_FILE)
-        self.lp_monitor = file.monitor_file(Gio.FileMonitorFlags.NONE)
-
-        # connect to signals
-        self.handlers[self.goa_auth] = \
-            self.goa_auth.connect('notify', lambda o, p: self.on_goa_auth_changed())
-        self.handlers[self.checkbutton_livepatch] = \
-            self.checkbutton_livepatch.connect('toggled', self.on_checkbutton_livepatch_toggled)
-        self.handlers[self.button_ubuntuone] =  \
-            self.button_ubuntuone.connect('clicked', self.on_button_ubuntuone_clicked)
-        self.handlers[self.lp_monitor] = \
-            self.lp_monitor.connect('changed', self.on_livepatch_status_changed)
-
-    def has_online_accounts(self):
-        try:
-            d = Gio.DesktopAppInfo.new('gnome-online-accounts-panel.desktop')
-            return d != None
-        except Exception:
-            return False
-
-    def is_livepatch_supported(self):
-        distro = aptsources.distro.get_distro()
-        di = distro_info.UbuntuDistroInfo()
-
-        return self.has_online_accounts() and \
-               di.is_lts(distro.codename) and \
-               distro.codename in di.supported(datetime.now().date())
-
-    def on_goa_auth_changed(self):
-        if self.goa_auth.logged:
-            self.button_ubuntuone.set_label(_('Sign Out'))
-
-            if self.goa_auth.token:
-                self.checkbutton_livepatch.set_sensitive(True)
-                self.label_livepatch_login.set_label(_('Signed in as %s' % self.goa_auth.username))
-            else:
-                self.checkbutton_livepatch.set_sensitive(False)
-                text = _('%s isn\'t authorized to use Livepatch.' % self.goa_auth.username)
-                text = "<span color='red'>" +  text + "</span>"
-                self.label_livepatch_login.set_markup(text)
-        else:
-            if self.is_livepatch_enabled() and not self.waiting_livepatch_response:
-                # Allow the user to disable livepatch even if
-                # the account expired (see LP: #1768797)
-                self.checkbutton_livepatch.set_sensitive(True)
-                self.label_livepatch_login.set_label(_('Livepatch is active.'))
-            else:
-                self.checkbutton_livepatch.set_sensitive(False)
-                self.label_livepatch_login.set_label(_('To use Livepatch you need to sign in.'))
-
-            self.button_ubuntuone.set_label(_('Sign In…'))
-
-    def on_livepatch_status_changed(self, file_monitor, file, other_file, event_type):
-        if not self.waiting_livepatch_response:
-            self.checkbutton_livepatch.set_active(self.is_livepatch_enabled())
-            self.on_goa_auth_changed()
-
-    def on_button_ubuntuone_clicked(self, button):
-        if self.goa_auth.logged:
-            self.do_logout()
-        else:
-            self.do_login()
-
-    def do_login(self):
-        try:
-            # Show login dialog!
-            dialog = DialogAuth(self.window_main, self.datadir)
-            response = dialog.run()
-        except Exception as e:
-            logging.error(e)
-            error(self.window_main,
-                  _("Error enabling Canonical Livepatch"),
-                  _("Please check your Internet connection."))
-        else:
-            if response == Gtk.ResponseType.OK:
-                self.goa_auth.login(dialog.account)
-                if self.goa_auth.logged:
-                    self.checkbutton_livepatch.set_active(True)
-
-    def do_logout(self):
-        self.checkbutton_livepatch.set_active(False)
-        self.goa_auth.logout()
-
-    def on_checkbutton_livepatch_toggled(self, checkbutton):
-        if self.waiting_livepatch_response:
-            return
-
-        self.waiting_livepatch_response = True
-
-        token = ''
-        enabled = False
-        if self.checkbutton_livepatch.get_active():
-            enabled = True
-            token = self.goa_auth.token if self.goa_auth.token else ''
-        self.backend.SetLivepatchEnabled(enabled, token,
-                                         reply_handler=self.livepatch_enabled_reply_handler,
-                                         error_handler=self.livepatch_enabled_error_handler,
-                                         timeout=LIVEPATCH_TIMEOUT)
-
-    def livepatch_enabled_reply_handler(self, is_error, prompt):
-        self.sync_checkbutton_livepatch(is_error, prompt)
-
-    def livepatch_enabled_error_handler(self, e):
-        if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-            logging.error("Authentication canceled, changes have not been saved")
-            self.sync_checkbutton_livepatch(is_error=True, prompt=None)
-        else:
-            self.sync_checkbutton_livepatch(is_error=True, prompt=str(e))
-
-    def sync_checkbutton_livepatch(self, is_error, prompt):
-        if is_error:
-            self.waiting_livepatch_response = False
-            self.checkbutton_livepatch.handler_block(self.handlers[self.checkbutton_livepatch])
-            self.checkbutton_livepatch.set_active(self.is_livepatch_enabled())
-            self.checkbutton_livepatch.handler_unblock(self.handlers[self.checkbutton_livepatch])
-
-            if prompt:
-                dialog = DialogLivepatchError(self.window_main, self.datadir)
-                response = dialog.run(prompt, show_settings_button=self.quit_when_livepatch_responds)
-                if response == DialogLivepatchError.RESPONSE_SETTINGS:
-                    self.window_main.show()
-                    self.quit_when_livepatch_responds = False
-        else:
-            do_dbus_call = False
-            if self.is_livepatch_enabled() and not self.checkbutton_livepatch.get_active():
-                do_dbus_call = True
-                enabled = False
-                token = ''
-            elif not self.is_livepatch_enabled() and self.checkbutton_livepatch.get_active():
-                do_dbus_call = True
-                enabled = True
-                token = self.goa_auth.token if self.goa_auth.token else ''
-            else:
-                self.waiting_livepatch_response = False
-
-            if do_dbus_call:
-                self.backend.SetLivepatchEnabled(enabled, token,
-                                                 reply_handler=self.livepatch_enabled_reply_handler,
-                                                 error_handler=self.livepatch_enabled_error_handler,
-                                                 timeout=LIVEPATCH_TIMEOUT)
-
-        self.on_goa_auth_changed()
-
-        if self.quit_when_livepatch_responds:
-            self.on_close_button(self.button_close)
